package javax.xml.stream;

import java.util.ServiceConfigurationError;
import java.security.AccessController;
import java.util.Iterator;
import java.util.ServiceLoader;
import java.security.PrivilegedAction;
import java.io.File;
import java.util.Properties;

class FactoryFinder
{
    private static final String DEFAULT_PACKAGE = "com.sun.xml.internal.";
    private static boolean debug;
    private static final Properties cacheProps;
    private static volatile boolean firstTime;
    private static final SecuritySupport ss;
    
    FactoryFinder() {
    }
    
    private static void dPrint(String msg) {
        if (FactoryFinder.debug) {
            System.err.println("JAXP: " + msg);
        }
    }
    
    private static Class getProviderClass(String className, ClassLoader cl, boolean doFallback, boolean useBSClsLoader) throws ClassNotFoundException {
        try {
            if (cl == null) {
                if (useBSClsLoader) {
                    return Class.forName(className, false, FactoryFinder.class.getClassLoader());
                }
                cl = FactoryFinder.ss.getContextClassLoader();
                if (cl == null) {
                    throw new ClassNotFoundException();
                }
                return Class.forName(className, false, cl);
            }
            else {
                return Class.forName(className, false, cl);
            }
        }
        catch (ClassNotFoundException e1) {
            if (doFallback) {
                return Class.forName(className, false, FactoryFinder.class.getClassLoader());
            }
            throw e1;
        }
    }
    
    static <T> T newInstance(Class<T> type, String className, ClassLoader cl, boolean doFallback) throws FactoryConfigurationError {
        return newInstance(type, className, cl, doFallback, false);
    }
    
    static <T> T newInstance(Class<T> type, String className, ClassLoader cl, boolean doFallback, boolean useBSClsLoader) throws FactoryConfigurationError {
        assert type != null;
        if (System.getSecurityManager() != null && className != null && className.startsWith("com.sun.xml.internal.")) {
            cl = null;
            useBSClsLoader = true;
        }
        try {
            Class<?> providerClass = getProviderClass(className, cl, doFallback, useBSClsLoader);
            if (!type.isAssignableFrom(providerClass)) {
                throw new ClassCastException(className + " cannot be cast to " + type.getName());
            }
            Object instance = providerClass.getConstructor(new Class[0]).newInstance(new Object[0]);
            if (FactoryFinder.debug) {
                dPrint("created new instance of " + providerClass + " using ClassLoader: " + cl);
            }
            return type.cast(instance);
        }
        catch (ClassNotFoundException x) {
            throw new FactoryConfigurationError("Provider " + className + " not found", x);
        }
        catch (Exception x2) {
            throw new FactoryConfigurationError("Provider " + className + " could not be instantiated: " + x2, x2);
        }
    }
    
    static <T> T find(Class<T> type, String fallbackClassName) throws FactoryConfigurationError {
        return find(type, type.getName(), null, fallbackClassName);
    }
    
    static <T> T find(Class<T> type, String factoryId, ClassLoader cl, String fallbackClassName) throws FactoryConfigurationError {
        dPrint("find factoryId =" + factoryId);
        try {
            String systemProp;
            if (type.getName().equals(factoryId)) {
                systemProp = FactoryFinder.ss.getSystemProperty(factoryId);
            }
            else {
                systemProp = System.getProperty(factoryId);
            }
            if (systemProp != null) {
                dPrint("found system property, value=" + systemProp);
                return newInstance(type, systemProp, cl, true);
            }
        }
        catch (SecurityException se) {
            throw new FactoryConfigurationError("Failed to read factoryId '" + factoryId + "'", se);
        }
        String configFile = null;
        try {
            if (FactoryFinder.firstTime) {
                synchronized (FactoryFinder.cacheProps) {
                    if (FactoryFinder.firstTime) {
                        configFile = FactoryFinder.ss.getSystemProperty("java.home") + File.separator + "lib" + File.separator + "stax.properties";
                        File f = new File(configFile);
                        FactoryFinder.firstTime = false;
                        if (FactoryFinder.ss.doesFileExist(f)) {
                            dPrint("Read properties file " + f);
                            FactoryFinder.cacheProps.load(FactoryFinder.ss.getFileInputStream(f));
                        }
                        else {
                            configFile = FactoryFinder.ss.getSystemProperty("java.home") + File.separator + "lib" + File.separator + "jaxp.properties";
                            f = new File(configFile);
                            if (FactoryFinder.ss.doesFileExist(f)) {
                                dPrint("Read properties file " + f);
                                FactoryFinder.cacheProps.load(FactoryFinder.ss.getFileInputStream(f));
                            }
                        }
                    }
                }
            }
            String factoryClassName = FactoryFinder.cacheProps.getProperty(factoryId);
            if (factoryClassName != null) {
                dPrint("found in " + configFile + " value=" + factoryClassName);
                return newInstance(type, factoryClassName, cl, true);
            }
        }
        catch (Exception ex) {
            if (FactoryFinder.debug) {
                ex.printStackTrace();
            }
        }
        if (type.getName().equals(factoryId)) {
            T provider = findServiceProvider(type, cl);
            if (provider != null) {
                return provider;
            }
        }
        else
            assert fallbackClassName == null;
        if (fallbackClassName == null) {
            throw new FactoryConfigurationError("Provider for " + factoryId + " cannot be found", null);
        }
        dPrint("loaded from fallback value: " + fallbackClassName);
        return newInstance(type, fallbackClassName, cl, true);
    }
    
    private static <T> T findServiceProvider(final Class<T> type, final ClassLoader cl) {
        try {
            return AccessController.doPrivileged(new PrivilegedAction<T>() {
                @Override
                public T run() {
                    ServiceLoader<T> serviceLoader;
                    if (cl == null) {
                        serviceLoader = ServiceLoader.load(type);
                    }
                    else {
                        serviceLoader = ServiceLoader.load(type, cl);
                    }
                    Iterator<T> iterator = serviceLoader.iterator();
                    if (iterator.hasNext()) {
                        return iterator.next();
                    }
                    return null;
                }
            });
        }
        catch (ServiceConfigurationError e) {
            RuntimeException x = new RuntimeException("Provider for " + type + " cannot be created", e);
            FactoryConfigurationError error = new FactoryConfigurationError(x, x.getMessage());
            throw error;
        }
    }
    
    static {
        FactoryFinder.debug = false;
        cacheProps = new Properties();
        FactoryFinder.firstTime = true;
        ss = new SecuritySupport();
        try {
            String val = FactoryFinder.ss.getSystemProperty("jaxp.debug");
            FactoryFinder.debug = (val != null && !"false".equals(val));
        }
        catch (SecurityException se) {
            FactoryFinder.debug = false;
        }
    }
}
